/****************************************************************************
** Playground related classes ( game engine etc ...)
**
**
** Copyright (C) 2001 Walter Rawdanik.  All rights reserved.
**
****************************************************************************/


#ifndef PLAYGROUND_H
#define PLAYGROUND_H

#include <qvector.h>
#include <qlist.h>
#include <qsortedlist.h>
#include <qbitarray.h>
#include <qpoint.h>
#include <qpixmap.h>
#include <qarray.h> 
#include <math.h>
#include "Puzzle.h"

class QPixmap;
class QCString;
class QPainter;

class Playblock
{
	
public:
	
	enum GAME_BLOCK
	{        
		RED,
		YELLOW,
		BLUE,
		GREEN,
		PURPLE,
		STONE,
		BROKEN_STONE,
		MOVING_STONE,
		FIRE_STONE,
		TRAP_STONE,
		EMPTY,
		TRANSITIONAL,
        WHITE,
        CYAN,
        GREY,
		INVALID
	};
	
	Playblock(int bx, int by,int x1=-1,int y1=-1, int x2=-9999,int y2=-9999, int aframe=0,GAME_BLOCK type=EMPTY,bool sel=false, bool grab=false);
	

	void recalculatePos(int nx, int ny);
	void recalculatePos();
   void createBlendedImage(QPixmap *tile, QPixmap *back, int ox, int oy, QPixmap *f);
   static QImage& blend(QImage& src, QImage& dst, float opacity);
   
   

	inline bool attached()
	{
		return iat;
	}
	
	inline void setLastDirection(int d)
	{
		iDir=d;
	}

	inline int lastDirection()
	{
		return iDir;
	}

	inline void setAttached(bool f=true)
	{
		iat=f;
	}

	inline void updatePos()
	{
		iy+=iay;
		ix+=iax;
	}

	inline void setX(int x)
	{
		ix=x;
	}
	inline void setY(int y)
	{
		iy=y;
	}

	inline void setAccelY(float a)
	{
		iay=a;
	}

	inline float accelY()
	{
		return iay;
	}

	inline void setAccelX(float a)
	{
		iax=a;
	}

	inline float accelX()
	{
		return iax;
	}


	inline int x()
	{
		return (int) (ix+0.5);
	}
	inline int y()
	{
		return (int)(iy+0.5);
	}
	inline int frame()
	{
		return iaf;
	}
	inline void setFrame(int f)
	{
		iaf=f;
	}

	inline void setStartPos(int x1, int y1)
	{
		ix1=x1;
		iy1=y1;
	}
	inline void setEndPos(int x2, int y2)
	{
		ix2=x2;
		iy2=y2;
	}

	inline void setCounter(int c)
	{
		cnt=c;
	}
	inline int counter()
	{
		return cnt;
	}

	inline bool isPointInside(int y)
	{	
		if(y>=iy && y<iy+Puzzle::bSize)
			return true;
		else
			return false;
	}


	inline int bx()
	{
		return cx;
	}

	inline int by()
	{
		return cy;
	}

	inline void setCurrentBlockPos(int bx, int by)
	{
		cx=bx;
		cy=by;
	}

	inline int x1()
	{
		return ix1;
	}
	inline int y1()
	{
		return iy1;
	}
	inline int x2()
	{
		return ix2;
	}
	inline int y2()
	{
		return iy2;
	}

	inline GAME_BLOCK type()
	{
		return itype;
	}

	inline void setType(GAME_BLOCK t)
	{
		itype=t;
	}
	
	
	inline void setSelected(bool f)
	{
		iSel=f;
	}

	inline bool selected()
	{
		return iSel;
	}

	inline void setGrabbed( bool f)
	{
		iGrab=f;
	}

	inline bool grabbed()
	{
		return iGrab;
	}

	inline bool updated()
	{
		return iUpd;
	}

	inline void setUpdated(bool f)
	{
		iUpd=f;
	}

	void draw(QPainter *p,int ox=0,int oy=0);
	


// We need these operators for QSortedList.sort()

	friend inline int operator < (Playblock& left, Playblock& right)
	{
		if(left.iy>right.iy)
			return true;
		else
			return false;
	}
	friend inline int operator == (Playblock& left, Playblock& right)
	{
		if(left.iy==right.iy)
			return true;
		else
			return false;
	}

	QString toString();
	bool fromString(const QString &s);


private:

	GAME_BLOCK itype;			// type of block

	int	iSel;				// is selected ( used to paint selection box)
	int	iGrab;				// is grabbed ( used to paint selection box)
	int	iaf;					// current frame of animation
	float ix;					// pos x ( in pixels)
	float iy;					// pos y ( in pixels)
	int cx;						// current pos x ( in blocks)
	int cy;						// current pos y ( in blocks)
	int ix1;					// starting pos x ( in blocks , used by movable blocks) 
	int iy1;					// starting pos y ( in blocks , used by movable blocks) 
	int ix2;					// ending pos x ( in blocks , used by movable blocks) 
	int iy2;					// ending pos y ( in blocks , used for moving blocks)	
	float iay;					// vertical acceleration
	float iax;					// horizontal acceleration ( used only for initial animation)
	int iat;				// whetber this box is attached to moving column
	int cnt;				// this is a generic counter used for controling speed of animation etc .. ( 33/sec )
	int iUpd;				// if true, block has been already processes by the current update()
	int iDir;				// last direction (1=up 0 =down)
};



class Playground
{ 

public:	


	enum GAME_STATE
	{
		INVALID,			
		READY,
		ANIMATE,
		INPROGRESS,
		OVER,
		DEMO
	};

    enum GAME_TYPE
    {
        TIME_BASED,
        MOVES_BASED
    };


    Playground();
    ~Playground();

	bool loadLevel(const char * level[],unsigned int number=1);
	bool loadLevel(const QString & fileName,unsigned int number=1);

	QBitArray* paintPlayground(QPainter *p, int ox=0, int oy=0, bool drawAll=false);
	void paintAnimatedPlayground(QPainter *p, int ox=0, int oy=0);
	bool update();	


	bool setSelected(int px, int py, bool keyBoardMode=false);

   // return current positon of selected block ( in pixels, no offset, suitable to be used for setSelected - this is to emulate tapping with keyboard
   QPoint selectedPositon(int pd);

   inline GAME_TYPE gameType() const
   {
       return gType;
   }

   inline bool isDeletionInProgresss()
   {
       return (bool)deletionCounter;
   }

	inline int width()
	{
		return w;
	}
	inline int height()
	{
		return h;
	}
	inline int number()
	{
		return num;
	}
	inline int timeLimit()
	{
		return t;
	}
	inline int totalLevels()
	{
		return numLevels;
	}

	inline unsigned int currentPoints()
	{
		return levelPoints;
	}
	inline void clearPoints()
	{
		levelPoints=0;
	}

	void setGameState(GAME_STATE s);

	inline GAME_STATE gameState()
	{
		return gState;
	}

	inline bool isGameWon()
	{
		return gameWon;
	}
	inline bool isInfoChange()
	{
		bool result=updateInfoBar;
		updateInfoBar=false;
		return result;
	}

	inline int tickValue()
	{
		return tick;
	}
	
	bool savePlayground(const QString &fileName,unsigned int currPoints, const QString &levelFileName );
	bool loadPlayground(const QString &fileName, unsigned int *currPoints,const char * level[]);

	bool savedPlaygroundInfo(const QString &fileName,unsigned int *points, unsigned int *level);


private:

	bool moveBlock(int srcStripe, int destStripe, Playblock *pb);		// tries to move block into another stripe

	bool isEmpty(unsigned int bx, unsigned int by);		// returns if this position is empty
	inline bool isStone(unsigned int bx, unsigned int by)
	{
		Playblock::GAME_BLOCK b=grid.at(by*w+bx);
		return (b==Playblock::STONE);			
	}
	bool isEmptyByPixel( unsigned int px, unsigned int py);

	inline QRect pixelCoordinatesAt(unsigned  int bx, unsigned int by) // given block position , returns its pixel coordinates
	{
		return QRect(bx*Puzzle::bSize,by*Puzzle::bSize,
			Puzzle::bSize,Puzzle::bSize);		
	}

	inline QPoint blockCoordinatesAt(unsigned int px,unsigned int py)
	{
		return QPoint(px/Puzzle::bSize,py/Puzzle::bSize);
	}

	inline bool isBlockStatic(Playblock *pb)
	{
		QRect r=pixelCoordinatesAt(pb->bx(),pb->by());
		if(r.y()==pb->y())
			return true;
		else
			return false;

	}

	Playblock* getVBlock(unsigned int bx,unsigned int by);

	bool isMovingLimit(Playblock *pb);

   Playblock* applyAcceleration(int stripeNum,Playblock *pb, float ac, bool normalize=false);
	void  detachBlocks(int stripeNum,Playblock *pb);

	Playblock* blockAt(int bx, int by);	//returns block at block position x,y 


	void initGrid(int w,int h);

	bool parseLevelHeader(const char *line);
	bool parseLevelLine(const char *line, int cl);
	void parseAttachedBlocks();


	void handleStripe(int stripeNum);
	void handleMovingBlockUp( int stripeNum, Playblock *pb);
	void handleMovingBlockDown( int stripeNum, Playblock *pb);
	bool handleBlock(int stripeNum, Playblock *pb);
	bool handleBrokenBlock(int stripeNum, Playblock *pb);
	bool handleFireBlock(int stripeNum, Playblock *pb);
	void handleTrapBlock(int stripeNum, Playblock *pb);
	void handleDeletion();
	void handleAnimation();

	bool clearBlocks();

	bool clearSelected();


	//bool updateBlock(Playblock *pb);

   QArray<Playblock::GAME_BLOCK> grid;					// general outlay of the game grid ( 1- blocked, 0 - passable)
   QBitArray stripes;				// list of update stripes ( 1- needs to be update, 0 - no need to update)
   QVector< QSortedList<Playblock> > dBlocks;	// holds lists of movable objects for each stripe
   QVector< Playblock > vBlocks;	// used to create 2d grid of clearable blocks for proximity checking and also for initial animation
   QList< Playblock > sBlocks;	// list of special blocks ( fire, trap etc ..)
   unsigned int w;					// width of the level ( in blocks)
   unsigned int h;					// height of the level ( in blocks)
   unsigned int num;				// level number
   unsigned int t;					// time alloted for level completion (in seconds)   
   Playblock *empty;
   Playblock *stone;
   Playblock *sb;					// currently selected block

   int	deletionCounter;			// if > 0 then we are stuck in deletion loop
   bool needsClearCheck;			// if true, check for blocks proximity needs to be performed	
   int beat;						// clock beat( 1 per second )
   unsigned int levelPoints;				// total # of points 
   bool updateInfoBar;				// if true, points or time needs to be updated
   bool gameWon;					// after level is over, returns true if it was sucessfully completed	
   GAME_STATE gState;
   int totalNumBlocks;				// if ever gets to zero, the level has been solved
   int numLevels;
   int tick;						// internal ever-increasing clock
   int initTicks;
   GAME_TYPE gType;                 // game type 
   	
};
#endif // PLAYGROUND_H

